# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-07
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import copy
import numpy
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\t'.join(out))


def _get_name_and_time_range(f_handler: netCDF4.Dataset,
                             input_path: str
                             ) -> typing.Tuple[str, str, str]:
    """"""
    _granule_name = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(_granule_name)
    start_str = f_handler.firstMeasurementTime
    stop_str = f_handler.lastMeasurementTime

    # L2 NetCDF filename and first/lastMeasurementTime attributes are not
    # reliable on older versions of the IPF, so acquire the info from the name
    # of the source SAFE instead
    _ipf_version = getattr(f_handler, 'IPFversion', 0.0)
    ipf_version = float(_ipf_version)
    if ipf_version < 3.4:
        source_product = f_handler.sourceProduct
        source_split = source_product.split('_')
        start_str = source_split[6]
        start_dt = datetime.datetime.strptime(start_str, '%Y%m%dT%H%M%S')
        start_str = f'{start_dt:%Y-%m-%dT%H:%M:%SZ}'
        stop_str = source_split[7]
        stop_dt = datetime.datetime.strptime(stop_str, '%Y%m%dT%H%M%S')
        stop_str = f'{stop_dt:%Y-%m-%dT%H:%M:%SZ}'
        unique_str = source_split[-1].split('.')[0]
        granule_name = f'{granule_name}_{unique_str}'

    return granule_name, start_str, stop_str


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables: typing.Dict[str, typing.Dict[str, typing.Any]]
    variables = {'rvlRadVel': {'name': 'rvlRadVel',
                               'options': {}},
                 'rvlRadVelStd': {'name': 'rvlRadVelStd',
                                  'options': {}},
                 'rvlUssX': {'name': 'rvlUssX',
                             'options': {}},
                 'rvlUssY': {'name': 'rvlUssY',
                             'options': {}},
                 'rvlSnr': {'name': 'rvlSnr',
                            'options': {}},
                 'rvlHeading': {},
                 'rvlIncidenceAngle': {},
                 'rvlSweepAngle': {}}

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read variables
    f_handler = netCDF4.Dataset(input_path, 'r')
    for var_id in variables.keys():
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        _band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = numpy.ma.array(_band)
    lon = idf_converter.lib.extract_variable_values(f_handler, 'rvlLon')
    lat = idf_converter.lib.extract_variable_values(f_handler, 'rvlLat')
    _granule_name, start_str, stop_str = _get_name_and_time_range(f_handler,
                                                                  input_path)
    # TODO: check if we should use rvlZeroDopplerTime instead
    _platform = f_handler.missionName
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)
    nsubswath = 0
    has_subswath = 'rvlSwath' in f_handler.dimensions
    if has_subswath:
        nsubswath = f_handler.dimensions['rvlSwath'].size
    f_handler.close()

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    if 0 >= nsubswath:
        for key in granule.vars:
            fake_subswath = granule.vars[key]['array'][:, :, numpy.newaxis]
            granule.vars[key]['array'] = fake_subswath
        nsubswath = 1

    if '.' in start_str:
        start_dt = datetime.datetime.strptime(start_str,
                                              '%Y-%m-%dT%H:%M:%S.%fZ')
    else:
        start_dt = datetime.datetime.strptime(start_str, '%Y-%m-%dT%H:%M:%SZ')

    if '.' in stop_str:
        stop_dt = datetime.datetime.strptime(stop_str,
                                             '%Y-%m-%dT%H:%M:%S.%fZ')
    else:
        stop_dt = datetime.datetime.strptime(stop_str, '%Y-%m-%dT%H:%M:%SZ')
    platform = f'Sentinel-1{_platform[-1].upper()}'

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 1000  # TODO: check this
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = platform
    granule.meta['sensor'] = 'C-SAR'

    # Include the L2 SAFE name as a global attribute if the input file was in
    # a SAFE directory layout
    input_dir = os.path.dirname(input_path)
    parent_name = os.path.basename(input_dir)
    if 'measurement' == parent_name:
        grandparent_name = os.path.basename(os.path.dirname(input_dir))
        if grandparent_name.endswith('.SAFE'):
            granule.meta['L2_SAFE'] = grandparent_name

    for subswath in range(0, nsubswath):

        # Replicate information shared by all subgranules
        subgranule = idf_converter.lib.create_granule(granule.idf_version,
                                                      granule.data_model)
        subgranule.meta = copy.deepcopy(granule.meta)
        subgranule.dims = copy.deepcopy(granule.dims)

        subgranule.meta['idf_granule_id'] = f'{_granule_name}_{subswath}'

        # Make sure to keep only slices where geographical coordinates are
        # fully defined
        sublon = granule.vars['lon']['array'][:, :, subswath]
        sublat = granule.vars['lat']['array'][:, :, subswath]
        valid_mask = (~numpy.ma.getmaskarray(sublon) &
                      ~numpy.ma.getmaskarray(sublat))
        valid_ind = numpy.where(valid_mask)
        valid_slices = (slice(valid_ind[0].min(), valid_ind[0].max() + 1),
                        slice(valid_ind[1].min(), valid_ind[1].max() + 1),
                        subswath)

        # Adapt dimensions
        sub_nrow = 1 + valid_ind[0].max() - valid_ind[0].min()
        sub_ncell = 1 + valid_ind[1].max() - valid_ind[1].min()
        subgranule.dims['row'] = int(sub_nrow)
        subgranule.dims['cell'] = int(sub_ncell)

        # Adapt variables
        subvars: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        for var_id in granule.vars.keys():
            subvars[var_id] = {}
            for attr_name, attr_value in granule.vars[var_id].items():
                if 'array' == attr_name:
                    _values = attr_value[valid_slices]
                    subvars[var_id]['array'] = _values
                    subvars[var_id]['valid_min'] = numpy.nanmin(_values)
                    subvars[var_id]['valid_max'] = numpy.nanmax(_values)
                else:
                    subvars[var_id][attr_name] = attr_value
            subgranule.vars = subvars

        # Add transforms
        transforms: TransformsList = []

        transforms.append(('s1_experimental_radial_velocity',
                           {'targets': ('rvlRadVel',),
                            'outputs': ('rvlRadVelExp',),
                            'incidence_angle': 'rvlIncidenceAngle',
                            'sweep_angle': 'rvlSweepAngle'}))

        vel_vars = ('rvlRadVel', 'rvlRadVelExp')
        transforms.append(('radial_velocity_sign', {'targets': vel_vars,
                                                    'update_extrema': True,
                                                    'heading': 'rvlHeading'}))

        # Remove direction and modulus variables
        to_remove = ('rvlHeading', 'rvlSweepAngle', 'rvlIncidenceAngle')
        transforms.append(('remove_vars', {'targets': to_remove}))

        channels = ['rvlRadVel', 'rvlRadVelExp', 'rvlRadVelStd', 'rvlUssX',
                    'rvlUssY', 'rvlSnr']
        output_opts['__export'] = channels

        yield input_opts, output_opts, subgranule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2021-08-13
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'


class InputPathMissing(IOError):
    """"""
    pass


class MissingVariablesList(Exception):
    """Error raised when the "variables" input option has not been specified
    by the user."""
    pass


class VariableNotFound(Exception):
    """Error raised when a requested variable is not available in the source
    file."""

    def __init__(self, var_id: str) -> None:
        """"""
        self.var_id = var_id


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    _var_ids = input_opts.get('variables', None)
    if _var_ids is None:
        raise MissingVariablesList()
    var_ids = [x.strip() for x in _var_ids.split(',')]

    f_handler = netCDF4.Dataset(input_path, 'r')
    channels = []
    for var_id in var_ids:
        if var_id not in f_handler.variables.keys():
            raise VariableNotFound(var_id)
        else:
            channels.append(var_id)
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        _band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        _data_slices = [slice(0, 1, None),  # time
                        slice(None, None, None), slice(None, None, None)]
        if 'Layer' in f_handler.variables[var_id].dimensions:
            _data_slices = [slice(0, 1, None),  # time
                            slice(0, 1, None),  # depth
                            slice(None, None, None), slice(None, None, None)]
        data_slices = tuple(_data_slices)
        granule.vars[var_id]['array'] = numpy.ma.array(_band[data_slices])

        # Compute min/max
        band = granule.vars[var_id]['array']
        band_mask = numpy.ma.getmaskarray(band)
        if band_mask.all():
            granule.vars[var_id]['valid_min'] = 0.0
            granule.vars[var_id]['valid_max'] = 1.0
        else:
            finite_ind = numpy.where(numpy.isfinite(band))
            granule.vars[var_id]['valid_min'] = numpy.min(band[finite_ind])
            granule.vars[var_id]['valid_max'] = numpy.max(band[finite_ind])

    lon = idf_converter.lib.extract_variable_values(f_handler, 'Longitude')
    lat = idf_converter.lib.extract_variable_values(f_handler, 'Latitude')
    _time = idf_converter.lib.extract_variable_values(f_handler, 'MT')
    time_units = f_handler.variables['MT'].units
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)
    f_handler.close()

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)

    dt = netCDF4.num2date(_time[0], time_units)

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 9258  # 1/12°
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['idf_granule_id'] = f'{granule_name}_{dt:%Y%m%d%H%M%S}'
    granule.dims['row'] = lat.shape[0]
    granule.dims['cell'] = lon.shape[1]

    # Format spatial information
    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    start_dt = netCDF4.num2date(_time[0], time_units)
    stop_dt = start_dt + datetime.timedelta(hours=1)
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    # Add transforms
    transforms: TransformsList = []

    output_opts['__export'] = channels

    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2023-07-15
#
# Copyright (C) 2016-2023 OceanDataLab
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import logging
import netCDF4
import idf_converter.lib
from idf_converter.lib.types import ReaderResult, TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'


class InputPathMissing(IOError):
    """Error raised when the "path" input option has not been provided by the
    user."""
    pass


def build_gcps(lon: numpy.typing.NDArray,
               lat: numpy.typing.NDArray,
               subsampling: typing.Sequence[int]
               ) -> typing.Dict[str, numpy.typing.NDArray]:
    """"""
    if numpy.ma.is_masked(lon) or numpy.ma.is_masked(lat):
        logger.warn('Some lon and/or lat is masked')

    lon_shape = numpy.array(numpy.shape(lon))  # lat should have the same shape
    x = subsampling[1]
    ngcps_s = numpy.ceil(lon_shape / subsampling).astype('int')
    ngcps_d = numpy.ceil(lon_shape / subsampling).astype('int')

    len_ac = lon_shape[1] - 1
    mid_ac = int(len_ac / 2)
    gcppix = numpy.array([0 * x, 10 * x, 20 * x, 30 * x, mid_ac - 1,
                          mid_ac, mid_ac + 1, mid_ac + 10 * x, mid_ac + 20 * x,
                          mid_ac + 30 * x, len_ac])
    gcppix_s = gcppix
    ngcps_s[1] = len(gcppix_s)
    ngcps_d[1] = len(gcppix)

    gcplin_out = numpy.linspace(0, lon_shape[0] - 1,
                                num=ngcps_d[0]).round().astype('int32')
    gcplin_in = numpy.linspace(0, lon_shape[0] - 1,
                               num=ngcps_s[0]).round().astype('int32')

    gcps = {'geoloc_row': gcplin_in, 'geoloc_cell': gcppix,
            'input_row': gcplin_in, 'input_cell': gcppix,
            'output_row': gcplin_out, 'output_cell': gcppix}
    return gcps


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ''
    return ('\n'.join(inp), '\n'.join(out))


def read_data(input_opts: typing.Dict[str, typing.Any],
              output_opts: typing.Dict[str, typing.Any]
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'ssha': {'name': 'ssha',
                          'long_name': ('sea surface height above reference '
                                        'ellipsoid'),
                          'units': 'm',
                          'valid_min': -50.0,
                          'valid_max': 50.0,
                          'options': {}},
                 'ssha_unfiltered': {'name': 'ssha_unfiltered',
                                     'long_name': ('sea surface height above '
                                                   'reference ellipsoid'),
                                     'units': 'm',
                                     'valid_min': -50.0,
                                     'valid_max': 50.0,
                                     'options': {}},
                 'mdt': {'name': 'mdt',
                         'long_name': 'mean dynamic topography',
                         'units': 'm',
                         'valid_min': -50.0,
                         'valid_max': 50.0,
                         'options': {}},

                 'sigma0': {'name': 'sigma0',
                            'long_name': ('normalized radar cross section '
                                          '(sigma0)'),
                            'units': 'm',
                            'valid_min': -1000.0,
                            'valid_max': 10000.0,
                            'options': {}},
                 'dac': {'name': 'dac',
                         'long_name': 'Dynamic atmospheric correction',
                         'units': 'm',
                         'valid_min': -50.0,
                         'valid_max': 50.0,
                         'options': {}},
                 'mss': {'name': 'mss',
                         'long_name': 'Mean surface slope',
                         'units': 'm',
                         'valid_min': -50.0,
                         'valid_max': 50.0,
                         'options': {}},
                 'calibration': {'name': 'calibration',
                                 'long_name': ('Instrumental Calibration '
                                               '(xover)'),
                                 'units': 'm',
                                 'valid_min': -50.0,
                                 'valid_max': 50.0,
                                 'options': {}},
                 'ssha_noiseless': {'name': 'ssha_noiseless',
                                    'long_name': ('sea surface height anomaly '
                                                  'without noise'),
                                    'units': 'm',
                                    'valid_min': -50.0,
                                    'valid_max': 50.0,
                                    'options': {}},
                 'ssha_filtered': {'name': 'ssha_filtered',
                                   'long_name': ('sea surface height anomaly '
                                                 'without noise'),
                                   'units': 'm',
                                   'valid_min': -50.0,
                                   'valid_max': 50.0,
                                   'options': {}},
                 'ssha_unedited': {'name': 'ssha_unedited',
                                   'long_name': 'sea surface height anomaly',
                                   'units': 'm',
                                   'valid_min': -50.0,
                                   'valid_max': 50.0,
                                   'options': {}},
                 'quality_flag': {'name': 'quality_flag',
                                  'long_name': 'Quality flag',
                                  'units': 'm',
                                  'valid_min': 0.0,
                                  'valid_max': 1.0,
                                  'options': {}},
                 'ocean_tide': {'name': 'ocean_tide',
                                'long_name': 'Ocean tide',
                                'units': 'm',
                                'valid_min': -50.0,
                                'valid_max': 50.0,
                                'options': {}},
                 'ugos': {'name': 'ugos',
                          'long_name': ('Geostrophic Eastward component '
                                        'derived from SSH'),
                          'units': 'm',
                          'valid_min': -50.0,
                          'valid_max': 50.0,
                          'options': {}},
                 'ugos_filtered': {'name': 'ugos_filtered',
                                   'long_name': ('Geostrophic Eastward '
                                                 'component derived from SSH'),
                                   'units': 'm',
                                   'valid_min': -50.0,
                                   'valid_max': 50.0,
                                   'options': {}},
                 'ugosa': {'name': 'ugosa',
                           'long_name': ('Geostrophic Eastward component '
                                         'derived from SSH Anomaly'),
                           'units': 'm',
                           'valid_min': -50.0,
                           'valid_max': 50.0,
                           'options': {}},
                 'ugosa_filtered': {'name': 'ugosa_filtered',
                                    'long_name': ('Geostrophic Eastward '
                                                  'component derived from SSH '
                                                  'Anomaly'),
                                    'units': 'm',
                                    'valid_min': -50.0,
                                    'valid_max': 50.0,
                                    'options': {}},
                 'vgos': {'name': 'vgos',
                          'long_name': ('Geostrophic Northward component '
                                        'derived from SSH'),
                          'units': 'm',
                          'valid_min': -50.0,
                          'valid_max': 50.0,
                          'options': {}},
                 'vgos_filtered': {'name': 'vgos_filtered',
                                   'long_name': ('Geostrophic Northward '
                                                 'component derived from SSH'),
                                   'units': 'm',
                                   'valid_min': -50.0,
                                   'valid_max': 50.0,
                                   'options': {}},
                 'vgosa': {'name': 'vgosa',
                           'long_name': ('Geostrophic Northward component '
                                         'derived from SSH Anomaly'),
                           'units': 'm',
                           'valid_min': -50.0,
                           'valid_max': 50.0,
                           'options': {}},
                 'vgosa_filtered': {'name': 'vgosa_filtered',
                                    'long_name': ('Geostrophic Northward '
                                                  'component derived from SSH '
                                                  'Anomaly'),
                                    'units': 'm',
                                    'valid_min': -50.0,
                                    'valid_max': 50.0,
                                    'options': {}},
                 'ssha_unedited_nolr': {'name': 'ssha_unedited_nolr',
                                        'long_name': ('sea surface height '
                                                      'above reference '
                                                      'ellipsoid 2, low '
                                                      'resolution removed'),
                                        'units': 'm',
                                        'valid_min': -50.0,
                                        'valid_max': 50.0,
                                        'options': {}},
                 'roughness': {'name': 'roughness',
                               'long_name': ('sea surface roughness from '
                                             'sigma0'),
                               'units': '',
                               'valid_min': -1000.0,
                               'valid_max': 1000.0,
                               'options': {}},

                 }

    _var_ids = input_opts.get('variables', None)
    if _var_ids is None:
        var_ids = list(variables.keys())
    else:
        var_ids = [x.strip() for x in _var_ids.split(',')]

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    idf_converter.lib.apply_global_overrides(input_opts, granule)
    idf_converter.lib.apply_var_overrides(input_opts, granule)

    maskable_channels = [_ for _ in var_ids if 'quality_flag' != _]
    granule.vars = {_: variables[_] for _ in var_ids if _ in variables.keys()}
    channels = list(granule.vars.keys())

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)
    filename = os.path.basename(input_path)
    subsampling = [40, 1]
    resolution = 2000
    if '_Unsmoothed_' in filename:
        subsampling = [200, 5]
        resolution = 250
    qf_mask: typing.Optional[numpy.typing.NDArray] = None

    f_handler = netCDF4.Dataset(input_path, 'r')

    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)

    # Extract geo coordinates information
    lon = f_handler.variables['longitude'][:, :]
    lat = f_handler.variables['latitude'][:, :]

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon[numpy.where(lon < 0.0)] = lon[numpy.where(lon < 0.0)] + 360.0

    # Extract time coordinates information
    t = f_handler.variables['time'][:]
    t_unit = f_handler.variables['time'].units
    t_mask = numpy.ma.getmaskarray(t)
    valid_t_ind = numpy.where(~t_mask)[0]
    time = t[valid_t_ind]
    lon = lon[valid_t_ind, :]
    lat = lat[valid_t_ind, :]
    for var_id in channels:
        if 'roughness' == var_id:
            if 'sigma0' in channels:
                continue

            # Add sigma0 for future roughness computations
            var_id = 'sigma0'
            granule.vars[var_id] = variables[var_id]
        elif 'ssha_unedited_nolr' == var_id:
            if 'ssha_unedited' in channels:
                continue

            # Add ssha_unedited for future ssha_unedited_nolr computations
            var_id = 'ssha_unedited'
            granule.vars[var_id] = variables[var_id]

        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        values = f_handler.variables[var_id][valid_t_ind, :]
        granule.vars[var_id]['array'] = values

    if 'quality_flag' in f_handler.variables.keys():
        quality_flag = f_handler.variables['quality_flag'][:]
        qf = quality_flag[valid_t_ind, :]
        qf_mask = (qf == 1)

    f_handler.close()

    nrow, ncell = numpy.shape(lon)
    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    start_dt = netCDF4.num2date(time[0], t_unit)
    stop_dt = netCDF4.num2date(time[-1], t_unit)

    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = resolution
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    # Add transforms
    transforms: TransformsList = []

    if 'roughness' in channels:
        transforms.append(('compute_swot_roughness',
                           {'targets': ('roughness', )}))

    if 'ssha_unedited_nolr' in channels:
        transforms.append(('compute_swot_ssha_unedited_nolr',
                           {'targets': ('ssha_unedited_nolr', )}))

    if qf_mask is not None:
        transforms.append(('static_common_mask', {'targets': maskable_channels,
                                                  'mask': qf_mask}))
    # - Remove variables
    to_remove = []
    if 'sigma0' in granule.vars and 'sigma0' not in channels:
        to_remove.append('sigma0')
    if 'ssha_unedited' in granule.vars and 'ssha_unedited' not in channels:
        to_remove.append('ssha_unedited')
    if len(to_remove) > 0:
        transforms.append(('remove_vars', {'targets': to_remove}))

    output_opts['__export'] = channels
    output_opts['gcp_distribution'] = build_gcps(lon, lat,
                                                 subsampling=subsampling)
    input_opts['geoloc_at_pixel_center'] = 'true'
    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module provides ."""

import numpy
import typing
import logging
import netCDF4
import pkg_resources
from idf_converter.lib.types import InputOptions, OutputOptions, ProcessedLists
from idf_converter.lib.types import Granule, FormatterResult

logger = logging.getLogger(__name__)


class IncompatibleNetCDFVersion(EnvironmentError):
    """Error raised when the version of the netCDF4 C library is too old and
    cannot create files that follow time-dependent IDF specifications."""
    def __init__(self, netcdf4_version: str) -> None:
        """"""
        self.netcdf4_version = netcdf4_version


def help() -> typing.Tuple[str, str]:
    """"""
    return ('', '')


def preformat(input_opts: InputOptions, output_opts: OutputOptions,
              granule: Granule
              ) -> typing.Iterator[FormatterResult]:
    """"""
    yield input_opts, output_opts, granule


def format_granule(input_opts: InputOptions,
                   output_opts: OutputOptions,
                   granule: Granule,
                   idf_file: netCDF4.Dataset) -> ProcessedLists:
    """Create dimensions, variables and attributes required for time-dependent
    granules.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.types.Granule
        Object representing the granule data and metadata
    idf_file: netCDF4.Dataset
        Handle of the IDF file to edit

    Returns
    -------
    tuple
        Information about granule information that the method has processed.
        The tuple contains three lists::
        - a list of names for the dimensions created by the method
        - a list of identifiers for the variables created by the method
        - a list of names for the attributes created by the method
    """
    # Check that the netCDF4 C library is at least 4.7.3 as a bug prevented the
    # creation of the time-dependent variables/dimensions scheme in earlier
    # versions
    _nc_version = netCDF4.__netcdf4libversion__
    nc_version = pkg_resources.parse_version(_nc_version.split('-', 1)[0])
    ref_version = pkg_resources.parse_version('4.7.3')
    if (nc_version < ref_version) and ('4.7.3-development' != _nc_version):
        logger.error('The netCDF C library available on the system is too old '
                     'and contains a bug that prevents the converter from '
                     'creating the output IDF file. This bug has been fixed '
                     'in version 4.7.3 so make sure to update the library at '
                     'least to this version.')
        raise IncompatibleNetCDFVersion(netCDF4.__netcdf4libversion__)

    # One GCP per time
    gcplon = granule.vars['lon']['array']
    gcplat = granule.vars['lat']['array']
    times = granule.vars['time']['array']

    npix = times.shape[0]
    gcpind = numpy.arange(npix)

    # Dimensions
    idf_file.createDimension('time', None)  # unlimited
    idf_file.createDimension('time_gcp', granule.dims['time'])
    idf_dims = ('time', 'time_gcp',)
    extra_dims = [x for x in granule.dims.keys() if x not in idf_dims]
    for extra_dim in extra_dims:
        idf_file.createDimension(extra_dim, granule.dims[extra_dim])

    # Variables
    # Dimension of the "time" variable should be "time_gcp" but there is a bug
    # in netCDF4 or hdf5 wherein a variable which shares the same name as an
    # unlimited dimension must also have an limited dimension (whereas
    # "time_gcp" is a finite dimension)
    _time = idf_file.createVariable('time', 'f8', ('time_gcp',))
    _time.long_name = 'time'
    _time.standard_name = 'time'
    _time.units = 'seconds since 1970-01-01T00:00:00.000000Z'
    _time.calendar = 'standard'
    _time.axis = 'T'
    _time[:] = times

    _latgcp = idf_file.createVariable('lat_gcp', 'f4', ('time_gcp',))
    _latgcp.long_name = 'ground control points latitude'
    _latgcp.standard_name = 'latitude'
    _latgcp.units = 'degrees_north'
    _latgcp.comment = 'geographical coordinates, WGS84 projection'
    _latgcp[:] = gcplat.astype('float32')

    _longcp = idf_file.createVariable('lon_gcp', 'f4', ('time_gcp',))
    _longcp.long_name = 'ground control points longitude'
    _longcp.standard_name = 'longitude'
    _longcp.units = 'degrees_east'
    _longcp.comment = 'geographical coordinates, WGS84 projection'
    _longcp[:] = gcplon.astype('float32')

    _indexdim1gcp = idf_file.createVariable('index_time_gcp', 'i4',
                                            ('time_gcp',))
    _indexdim1gcp[:] = gcpind.astype('int32')
    _indexdim1gcp.long_name = ('index of ground control points in time '
                               'dimension')
    _indexdim1gcp.comment = ('index goes from 0 (start of first pixel) to '
                             'dimension value (end of last pixel)')

    idf_file.cdm_data_type = b'Trajectory'

    processed_dims = ['time', 'time_gcp']
    processed_vars = ['time', 'lat_gcp', 'lon_gcp', 'index_time_gcp', 'lat',
                      'lon']
    processed_attrs = ['cdm_data_type']

    return (processed_dims, processed_vars, processed_attrs)
